package org.artifactory.util;

import lombok.experimental.UtilityClass;
import org.apache.commons.collections.MapUtils;
import org.apache.http.HttpStatus;
import org.apache.http.client.methods.CloseableHttpResponse;
import org.apache.http.client.methods.HttpGet;
import org.apache.http.impl.client.CloseableHttpClient;
import org.artifactory.api.rest.constant.RestConstants;
import org.artifactory.api.rest.constant.SystemRestConstants;
import org.artifactory.repo.http.CloseableHttpExecute;

import java.io.IOException;
import java.io.UncheckedIOException;
import java.util.Map;
import java.util.function.Function;

/**
 * Utility class to get the Artifactory service version via the new API with backward compatibility for older versions.
 *
 * @author Andrey Tashlitsky
 */
@UtilityClass
public class ArtifactoryVersionUtil {

    public CloseableHttpResponse getArtifactoryServiceVersion(CloseableHttpExecute closeableHttpExecute,
            String instanceUrl, Map<String, String> headers) {
        return getArtifactoryServiceVersionFunction(httpGet -> {
            try {
                return closeableHttpExecute.executeMethod(httpGet);
            } catch (IOException e) {
                throw new UncheckedIOException(e);
            }
        }, instanceUrl, headers);
    }

    public CloseableHttpResponse getArtifactoryServiceVersion(CloseableHttpClient closeableHttpClient,
            String instanceUrl, Map<String, String> headers) {
        return getArtifactoryServiceVersionFunction(httpGet -> {
            try {
                return closeableHttpClient.execute(httpGet);
            } catch (IOException e) {
                throw new UncheckedIOException(e);
            }
        }, instanceUrl, headers);
    }

    private CloseableHttpResponse getArtifactoryServiceVersionFunction(
            Function<HttpGet, CloseableHttpResponse> getServiceVersion, String instanceUrl,
            Map<String, String> headers) {
        String basePath = instanceUrl + "/" + RestConstants.PATH_API + "/" + SystemRestConstants.PATH_ROOT + "/";
        HttpGet serviceVersionHttpGet = new HttpGet(basePath + SystemRestConstants.PATH_VERSION_SERVICE);
        setHeaders(serviceVersionHttpGet, headers);
        return getServiceVersionWithBackwardCompatibility(getServiceVersion, headers, basePath, serviceVersionHttpGet);
    }

    private CloseableHttpResponse getServiceVersionWithBackwardCompatibility(
            Function<HttpGet, CloseableHttpResponse> getServiceVersion, Map<String, String> headers, String basePath,
            HttpGet serviceVersionHttpGet) {
        CloseableHttpResponse versionResponse = getServiceVersion.apply(serviceVersionHttpGet);
        if (versionResponse.getStatusLine().getStatusCode() == HttpStatus.SC_NOT_FOUND) {
            HttpGet oldVersionGet = new HttpGet(basePath + SystemRestConstants.PATH_VERSION);
            setHeaders(serviceVersionHttpGet, headers);
            return getServiceVersion.apply(oldVersionGet);
        }
        return versionResponse;
    }

    private void setHeaders(HttpGet httpGet, Map<String, String> headers) {
        if (MapUtils.isEmpty(headers)) {
            return;
        }
        for (Map.Entry<String, String> header : headers.entrySet()) {
            httpGet.addHeader(header.getKey(), header.getValue());
        }
    }
}
