package org.artifactory.ph;


import org.apache.commons.io.IOUtils;
import org.artifactory.common.resource.BaseResource;
import org.artifactory.core.api.ph.exception.PackageException;
import org.artifactory.core.api.ph.virtual.resolution.PackageMetadataRow;
import org.artifactory.core.api.ph.virtual.resolution.PackageRequestContext;
import org.artifactory.core.api.ph.virtual.resolution.PackageVirtualResolutionAdapter;
import org.artifactory.fs.RepoResource;
import org.artifactory.util.LazyStreamLoader;
import org.artifactory.util.StringInputStream;
import org.junit.jupiter.api.Assertions;
import org.junit.jupiter.api.Test;
import org.mockito.Mockito;
import org.testng.annotations.BeforeMethod;

import java.io.IOException;
import java.io.InputStream;
import java.nio.charset.StandardCharsets;
import java.util.LinkedHashSet;
import java.util.Map;

import static org.mockito.ArgumentMatchers.any;
import static org.mockito.Mockito.times;
import static org.mockito.Mockito.verify;
import static org.mockito.Mockito.when;
import static org.mockito.MockitoAnnotations.openMocks;

class PackageVirtualResolutionServiceTest {
    private final static String virtualRepoKey1 = "virtualRepoKey1";
    private final static String virtualRepoKey2 = "virtualRepoKey2";

    @BeforeMethod
    public void setUp() throws PackageException {
        openMocks(this);
    }

    @Test
    void shouldUsePackageVirtualResolution() {
        PackageVirtualResolutionAdapter<PackageMetadataRow> packageVirtualAdapter =
                Mockito.mock(PackageVirtualResolutionAdapter.class);
        PackageVirtualResolutionServiceImpl packageVirtualResolutionService =
                new PackageVirtualResolutionServiceImpl(packageVirtualAdapter);
        when(packageVirtualAdapter.shouldUsePackageVirtualResolution(virtualRepoKey1)).thenReturn(true);
        when(packageVirtualAdapter.shouldUsePackageVirtualResolution(virtualRepoKey2)).thenReturn(false);
        boolean result = packageVirtualResolutionService.
                shouldUsePackageVirtualResolution(virtualRepoKey1);
        Assertions.assertTrue(result);
        result = packageVirtualResolutionService.
                shouldUsePackageVirtualResolution(virtualRepoKey2);
        Assertions.assertFalse(result);
        verify(packageVirtualAdapter, times(2)).shouldUsePackageVirtualResolution(any());
    }

    @Test
    void getItemInfoFromVirtual() throws PackageException {
        PackageVirtualResolutionAdapter<PackageMetadataRow> packageVirtualAdapter =
                Mockito.mock(PackageVirtualResolutionAdapter.class);
        PackageVirtualResolutionServiceImpl packageVirtualResolutionService =
                new PackageVirtualResolutionServiceImpl(packageVirtualAdapter);
        RepoResource resource = Mockito.mock(RepoResource.class);
        when(resource.getPath()).thenReturn("test");
        when(packageVirtualAdapter.getItemInfoFromVirtual(any(), any())).thenReturn(resource);
        BaseResource res = packageVirtualResolutionService.getItemInfoFromVirtual(any(), any());
        Assertions.assertEquals("test", res.getPath());
        verify(packageVirtualAdapter, times(1)).getItemInfoFromVirtual(any(), any());
    }

    @Test
    void extract() throws PackageException {
        RepoResource resource = Mockito.mock(RepoResource.class);
        PackageVirtualResolutionAdapter packageVirtualAdapter =
                Mockito.mock(PackageVirtualResolutionAdapter.class);
        PackageVirtualResolutionServiceImpl packageVirtualResolutionService =
                new PackageVirtualResolutionServiceImpl(packageVirtualAdapter);
        Map<String, LinkedHashSet<PackageMetadataRow>> map = Mockito.mock(Map.class);
        when(packageVirtualAdapter.extract(resource, virtualRepoKey1, "",null )).thenReturn(map);
        Map<String, LinkedHashSet<PackageMetadataRow>> extracted = packageVirtualResolutionService.
                extract(resource, virtualRepoKey1, "",null );
        Assertions.assertSame(extracted, map);
        verify(packageVirtualAdapter, times(1)).extract(any(), any(), any(),any() );
    }

    @Test
    void toInputStream() throws PackageException, IOException {
        PackageVirtualResolutionAdapter packageVirtualAdapter =
                Mockito.mock(PackageVirtualResolutionAdapter.class);
        PackageVirtualResolutionServiceImpl packageVirtualResolutionService =
                new PackageVirtualResolutionServiceImpl(packageVirtualAdapter);
        Map map = Mockito.mock(Map.class);
        PackageRequestContext context = Mockito.mock(PackageRequestContext.class);
        StringInputStream streamLoader = new StringInputStream("Test");
        when(packageVirtualAdapter.toInputStream(map, context)).thenReturn(new LazyStreamLoader((() -> streamLoader)));
        InputStream inputStream = packageVirtualResolutionService.toInputStream(map, context);
        Assertions.assertEquals("Test", IOUtils.toString(inputStream, StandardCharsets.UTF_8));
        verify(packageVirtualAdapter, times(1)).toInputStream(any(), any());
    }
}
