package org.artifactory.repo.db;

import org.artifactory.addon.AddonsManager;
import org.artifactory.addon.xray.XrayAddon;
import org.artifactory.api.context.ArtifactoryContext;
import org.artifactory.api.context.ArtifactoryContextThreadBinder;
import org.artifactory.api.security.AuthorizationService;
import org.artifactory.common.StatusHolder;
import org.artifactory.descriptor.repo.LocalRepoDescriptor;
import org.artifactory.descriptor.repo.RepoType;
import org.artifactory.repo.InternalRepoPathFactory;
import org.artifactory.repo.RepoPath;
import org.artifactory.repo.snapshot.MavenSnapshotVersionAdapter;
import org.artifactory.request.Request;
import org.artifactory.request.RequestContext;
import org.artifactory.spring.InternalArtifactoryContext;
import org.testng.annotations.Test;

import java.util.Arrays;
import java.util.Enumeration;

import static org.mockito.ArgumentMatchers.any;
import static org.mockito.ArgumentMatchers.eq;
import static org.mockito.Mockito.mock;
import static org.mockito.Mockito.when;
import static org.testng.Assert.assertEquals;

public class DbRepoDockerManifestDownloadTest {

    private AddonsManager addonsManager;

    @Test
    public void downloadUnscannedManifestAllowedOnlyForDockerRepoWhenHeaderSetToTrue() {
        Arrays.stream(RepoType.values()).forEach(repoType -> verifyDownloadAllowed(
                repoType,
                true,
                false,
                "X-JFrog-Bypass-Unscanned-Blocking",
                "true",
                "manifest.json",
                repoType.equals(RepoType.Docker)));
    }

    @Test
    public void downloadWhenHeaderSetToFalse() {
        Arrays.stream(RepoType.values()).forEach(repoType -> verifyDownloadAllowed(
                repoType,
                true,
                false,
                "X-JFrog-Bypass-Unscanned-Blocking",
                "false",
                "manifest.json",
                false));

        Arrays.stream(RepoType.values()).forEach(repoType -> verifyDownloadAllowed(
                repoType,
                false,
                false,
                "X-JFrog-Bypass-Unscanned-Blocking",
                "false",
                "manifest.json",
                true));

    }

    @Test
    public void downloadUnscannedManifestWithWrongHeader() {
        Arrays.stream(RepoType.values()).forEach(repoType -> verifyDownloadAllowed(
                repoType,
                true,
                false,
                "X-JFrog-Bypass-Download-Wrong",
                "false",
                "manifest.json",
                false));

        Arrays.stream(RepoType.values()).forEach(repoType -> verifyDownloadAllowed(
                repoType,
                false,
                false,
                "X-JFrog-Bypass-Download-Wrong",
                "false",
                "manifest.json",
                true
        ));

    }

    @Test
    public void downloadUnscannedFileWhenHeaderSetToTrue() {
        Arrays.stream(RepoType.values()).forEach(repoType -> verifyDownloadAllowed(
                repoType,
                true,
                false,
                "X-JFrog-Bypass-Unscanned-Blocking",
                "true",
                "other.file",
                false));


        Arrays.stream(RepoType.values()).forEach(repoType -> verifyDownloadAllowed(
                repoType,
                false,
                false,
                "X-JFrog-Bypass-Unscanned-Blocking",
                "true",
                "other.file",
                true));

    }

    @Test
    public void downloadManifestWhenRequestIsNotSet() {
        Arrays.stream(RepoType.values()).forEach(repoType -> verifyDownloadAllowed(
                repoType,
                true,
                true,
                "",
                "",
                "manifest.json",
                false));


        Arrays.stream(RepoType.values()).forEach(repoType -> verifyDownloadAllowed(
                repoType,
                false,
                true,
                "",
                "",
                "manifest.json",
                true));

    }

    public void verifyDownloadAllowed(
            RepoType repoType,
            boolean downloadBlockedByXRay,
            boolean requestIsNotSet,
            String headerName,
            String headerValue,
            String fileName,
            boolean downloadAllowed
    ) {
        setup();

        XrayAddon xrayAddon = mock(XrayAddon.class);
        when(addonsManager.addonByType(XrayAddon.class)).thenReturn(xrayAddon);
        when(xrayAddon.isDownloadBlocked(any(RepoPath.class), eq(true))).thenReturn(false);
        when(xrayAddon.isDownloadBlocked(any(RepoPath.class), eq(false))).thenReturn(downloadBlockedByXRay);
        when(xrayAddon.isXrayAliveByHeartbeat()).thenReturn(true);
        RequestContext requestContext = mock(RequestContext.class);
        if (requestIsNotSet) {
            when(requestContext.getRequest()).thenReturn(null);
        } else {
            Request request = mock(Request.class);
            Enumeration<String> enumeration = mock(Enumeration.class);
            if (headerName.isEmpty()) {
                when(enumeration.hasMoreElements()).thenReturn(false);
            } else {
                when(enumeration.hasMoreElements()).thenReturn(true);
                when(enumeration.nextElement()).thenReturn(headerValue);
                when(request.getHeaderValues(headerName)).thenReturn(enumeration);
                when(requestContext.getRequest()).thenReturn(request);
            }
        }
        RepoPath path = InternalRepoPathFactory.create("target", fileName);
        DbLocalRepo<LocalRepoDescriptor> repo = createDbRepo(repoType);
        StatusHolder status = repo.checkDownloadIsNotBlocked(requestContext, path);

        assertEquals(!status.isError(), downloadAllowed);

        cleanUp();
    }

    private void setup() {
        ArtifactoryContext context = mock(InternalArtifactoryContext.class);
        AuthorizationService authService = mock(AuthorizationService.class);
        addonsManager = mock(AddonsManager.class);
        when(context.beanForType(AddonsManager.class)).thenReturn(addonsManager);
        when(context.getAuthorizationService()).thenReturn(authService);
        ArtifactoryContextThreadBinder.bind(context);
    }

    private void cleanUp() {
        ArtifactoryContextThreadBinder.unbind();
    }

    private DbLocalRepo<LocalRepoDescriptor> createDbRepo(RepoType repoType) {
        LocalRepoDescriptor descriptor = new LocalRepoDescriptor();
        descriptor.setType(repoType);
        return new DbLocalRepo<>(descriptor, null, null) {

            @Override
            public MavenSnapshotVersionAdapter getMavenSnapshotVersionAdapter() {
                return null;
            }

            @Override
            public boolean isSuppressPomConsistencyChecks() {
                return false;
            }
        };
    }

}
