package org.artifactory.traffic;

import org.artifactory.security.access.AccessTokenAuthentication;
import org.artifactory.traffic.v2.TrafficEntryV2;
import org.jfrog.access.common.ServiceId;
import org.jfrog.access.common.ServiceType;
import org.jfrog.access.token.JwtAccessToken;
import org.springframework.security.core.Authentication;
import org.testng.annotations.DataProvider;
import org.testng.annotations.Test;

import java.util.UUID;

import static org.mockito.Mockito.doReturn;
import static org.mockito.Mockito.mock;
import static org.mockito.Mockito.when;
import static org.testng.Assert.assertEquals;

public class TrafficUtilsTest {

    @DataProvider(name = "subjects")
    public static Object[][] subjects() {
        return new Object[][] {
                {null, TrafficEntryV2.TrafficActor.USER},
                {"", TrafficEntryV2.TrafficActor.USER},
                {"@", TrafficEntryV2.TrafficActor.USER},
                {"jfxr@", TrafficEntryV2.TrafficActor.USER},
                {"@sded", TrafficEntryV2.TrafficActor.USER},
                {getFormattedName(ServiceType.ACCESS), TrafficEntryV2.TrafficActor.USER},
                {getFormattedName(ServiceType.ACCESS_LEGACY), TrafficEntryV2.TrafficActor.USER},
                {getFormattedName(ServiceType.ARTIFACTORY), TrafficEntryV2.TrafficActor.USER},
                {getFormattedName(ServiceType.ARTIFACTORY_LEGACY), TrafficEntryV2.TrafficActor.USER},
                {getFormattedName(ServiceType.MISSION_CONTROL), TrafficEntryV2.TrafficActor.USER},
                {getFormattedName(ServiceType.XRAY), TrafficEntryV2.TrafficActor.XRAY},
                {getFormattedName(ServiceType.XRAY_INDEXER), TrafficEntryV2.TrafficActor.XRAY},
                {getFormattedName(ServiceType.XRAY_PERSIST), TrafficEntryV2.TrafficActor.XRAY},
                {getFormattedName(ServiceType.XRAY_ANALYSIS), TrafficEntryV2.TrafficActor.XRAY},
                {getFormattedName(ServiceType.DISTRIBUTION), TrafficEntryV2.TrafficActor.USER},
                {getFormattedName(ServiceType.FRONTEND), TrafficEntryV2.TrafficActor.USER},
                {getFormattedName(ServiceType.EVENT), TrafficEntryV2.TrafficActor.USER},
                {getFormattedName(ServiceType.PIPELINES), TrafficEntryV2.TrafficActor.PIPELINES},
                {getFormattedName(ServiceType.REPLICATOR), TrafficEntryV2.TrafficActor.REPLICATOR},
                {getFormattedName(ServiceType.OBSERVABILITY), TrafficEntryV2.TrafficActor.OBSERVABILITY},
                {"jfxr@cve_applicability_12345678", TrafficEntryV2.TrafficActor.XRAY_REPETITIVE},
                {"jfxr@extra_analysis_12345678", TrafficEntryV2.TrafficActor.XRAY_REPETITIVE}
        };
    }

    @Test(dataProvider = "subjects")
    public void testAccessToken(String subject, TrafficEntryV2.TrafficActor expected) {
        AccessTokenAuthentication authentication = mock(AccessTokenAuthentication.class);
        JwtAccessToken token = mock(JwtAccessToken.class);
        when(token.getSubject()).thenReturn(subject);
        when(authentication.getAccessToken()).thenReturn(token);
        TrafficEntryV2.TrafficActor actual = TrafficUtils.getTrafficActor(null, null, authentication);
        assertEquals(expected, actual);
    }

    @DataProvider(name = "userdata")
    public static Object[][] userdata() {
        return new Object[][] {
                {"", "", TrafficEntryV2.TrafficActor.USER},
                {null, null, TrafficEntryV2.TrafficActor.USER},
                {"", "xray", TrafficEntryV2.TrafficActor.XRAY},
                {"useragentxray", "", TrafficEntryV2.TrafficActor.XRAY},
                {"useragentjfrog-replicator/", "", TrafficEntryV2.TrafficActor.REPLICATOR},
                {"observability/", "_system_", TrafficEntryV2.TrafficActor.OBSERVABILITY},
                {"observability/", "", TrafficEntryV2.TrafficActor.USER},
                {"jfconnect/", "_system_", TrafficEntryV2.TrafficActor.JFCONNECT},
                {"jfconnect/", "", TrafficEntryV2.TrafficActor.USER},
                {"ba-billing-logshipping", "_system_", TrafficEntryV2.TrafficActor.BA_BILLING_LOGSHIPPING},
                {"ba-billing-logshipping/", "_system_", TrafficEntryV2.TrafficActor.BA_BILLING_LOGSHIPPING},
                {"ba-billing-logshipping/", "", TrafficEntryV2.TrafficActor.USER},
                {"noise", "", TrafficEntryV2.TrafficActor.USER},
        };
    }

    @Test(dataProvider = "userdata")
    public void testUserData(String userAgent, String username, TrafficEntryV2.TrafficActor expected) {
        TrafficEntryV2.TrafficActor actual = TrafficUtils.getTrafficActor(userAgent, username, null);
        assertEquals(expected, actual);
    }

    @DataProvider
    public static Object[][] testGetTrafficActorV2DataProvider() {
        AccessTokenAuthentication jfxrToken = mock(AccessTokenAuthentication.class);
        JwtAccessToken jwtJfxrToken = mock(JwtAccessToken.class);
        doReturn(jwtJfxrToken).when(jfxrToken).getAccessToken();
        doReturn("jfxr@cool").when(jwtJfxrToken).getSubject();

        AccessTokenAuthentication jfpipToken = mock(AccessTokenAuthentication.class);
        JwtAccessToken jwtJfpipToken = mock(JwtAccessToken.class);
        doReturn(jwtJfpipToken).when(jfpipToken).getAccessToken();
        doReturn("jfpip@cool").when(jwtJfpipToken).getSubject();

        AccessTokenAuthentication jfcoolToken = mock(AccessTokenAuthentication.class);
        JwtAccessToken jwtJfcoolToken = mock(JwtAccessToken.class);
        doReturn(jwtJfcoolToken).when(jfcoolToken).getAccessToken();
        doReturn("jfcool@cool").when(jwtJfcoolToken).getSubject(); // No such service

        return new Object[][] {
                {"", false, null, TrafficEntryV2.TrafficActor.USER},
                {"cool", false, null, TrafficEntryV2.TrafficActor.USER},
                {"cool", true, jfpipToken, TrafficEntryV2.TrafficActor.PIPELINES},
                {"cool", true, jfxrToken, TrafficEntryV2.TrafficActor.XRAY},
                {"cool", true, jfcoolToken, TrafficEntryV2.TrafficActor.XRAY},
                {"jfrog-replicator/", false, null, TrafficEntryV2.TrafficActor.REPLICATOR},
                {"observability", false, null, TrafficEntryV2.TrafficActor.OBSERVABILITY},
                {"observability/", false, null, TrafficEntryV2.TrafficActor.OBSERVABILITY},
                {"ba-billing-logshipping", false, null, TrafficEntryV2.TrafficActor.BA_BILLING_LOGSHIPPING},
                {"ba-billing-logshipping/", false, null, TrafficEntryV2.TrafficActor.BA_BILLING_LOGSHIPPING}
        };
    }

    @Test(dataProvider = "testGetTrafficActorV2DataProvider")
    public void testGetTrafficActorV2(String identifier, boolean xrayAuthenticationUser, Authentication authentication,
            TrafficEntryV2.TrafficActor expectedActor) {
        TrafficEntryV2.TrafficActor trafficActor =
                TrafficUtils.getTrafficActor(identifier, xrayAuthenticationUser, authentication);
        assertEquals(trafficActor, expectedActor);
    }

    private static String getFormattedName(String serviceName) {
        return new ServiceId(serviceName, UUID.randomUUID().toString()).getFormattedName();
    }
}