package org.artifactory.opentracing;

import org.artifactory.api.events.EventLogHintManager;

import static org.artifactory.event.EventHintType.AsyncThread;
import static org.artifactory.mirror.MirrorConstants.ASYNC_THREAD;

/**
 * A runnable decorator that starts a server-side tracing span, either as an independent span or as a child of an
 * existing
 * span
 *
 * @author Uriah Levy
 */
public class TraceableRunnableDecorator implements Runnable {
    private final ArtifactoryTracer artifactoryTracer;
    private Runnable runnable;
    private ServerTracingSpan currentSpan;

    public TraceableRunnableDecorator(Runnable runnable, ArtifactoryTracer artifactoryTracer, boolean inheritSpan) {
        this.runnable = runnable;
        this.artifactoryTracer = artifactoryTracer;
        this.currentSpan = inheritSpan ? artifactoryTracer.getCurrentThreadActiveSpan() : null;
    }

    @Override
    public void run() {
        try {
            EventLogHintManager.put(AsyncThread, ASYNC_THREAD);
            if (currentSpan != null) {
                // Set the current span as active for the current thread
                try (ServerTracingScope ignored = artifactoryTracer.activateSpan(currentSpan)) {
                    runnable.run();
                    return;
                }
            }
            // Start a new span that ends at the end of this invocation
            try (ServerTracingScope ignored = artifactoryTracer
                    .startSpanOnCurrentThread("Artifactory Async Invocation")) {
                runnable.run();
            }
        } finally {
            EventLogHintManager.clear();
        }
    }
}
