/*
 *
 * Copyright 2018 JFrog Ltd. All rights reserved.
 * JFROG PROPRIETARY/CONFIDENTIAL. Use is subject to license terms.
 */

package org.artifactory.core.api.ph.vcs.git;

import com.google.common.collect.Sets;
import org.apache.commons.io.IOUtils;
import org.apache.commons.lang.StringUtils;

import java.io.InputStream;
import java.util.Set;
import java.util.stream.Collectors;

/**
 * Represents a git repository tags and branches.
 * Follows the get command 'curl https://github.com/twbs/bootstrap.git/info/refs?service=git-upload-pack'
 *
 * @author Shay Yaakov
 */
public class GitRefs {

    public static final String REFS_FILENAME = "gitrefs";
    public Set<GitRef> tags = Sets.newLinkedHashSet();
    public Set<GitRef> branches = Sets.newLinkedHashSet();

    public boolean isEmpty() {
        return tags.isEmpty() && branches.isEmpty();
    }

    public boolean contains(String tagOrBranch) {
        for (GitRef tagRef : tags) {
            if (StringUtils.equalsIgnoreCase(tagRef.name, tagOrBranch)) {
                return true;
            }
        }
        for (GitRef branchRef : branches) {
            if (StringUtils.equalsIgnoreCase(branchRef.name, tagOrBranch)) {
                return true;
            }
        }
        return false;
    }

    public void merge(GitRefs other) {
        if (other == null) {
            return;
        }

        tags.addAll(other.tags);
        branches.addAll(other.branches);
    }

    public InputStream constructOriginalRefsStream() {
        final StringBuilder sb = new StringBuilder();
        for (GitRef tag : tags) {
            String commit =
                    StringUtils.isNotBlank(tag.commitId) ? tag.commitId : "0000000000000000000000000000000000000000";
            sb.append("0000").append(commit).append(" refs/tags/").append(tag.name).append("\n");
        }
        for (GitRef branch : branches) {
            String commit = StringUtils.isNotBlank(branch.commitId) ? branch.commitId :
                    "0000000000000000000000000000000000000000";
            sb.append("0000").append(commit).append(" refs/heads/").append(branch.name).append("\n");
        }

        return IOUtils.toInputStream(sb.toString());
    }

    public Iterable<String> tagsRefsStrings() {
        return refsString(tags);
    }

    public Iterable<String> branchesRefsStrings() {
        return refsString(branches);
    }

    private Iterable<String> refsString(Set<GitRef> setToConvert) {
        return setToConvert.stream().map(ref -> ref.name.replaceFirst("^[=vV]+", "")).collect(Collectors.toList());
    }
}
