package org.artifactory.core.api.ph.virtual;

import com.google.common.collect.Maps;
import org.apache.commons.io.IOUtils;
import org.artifactory.common.repo.PackageType;
import org.artifactory.common.resource.BaseResource;
import org.artifactory.confprop.ConfProps;
import org.artifactory.core.api.ph.exception.PackageException;
import org.artifactory.core.api.ph.model.PackageDownloadContext;
import org.artifactory.core.api.ph.resource.PackageUnfoundedResource;
import org.artifactory.core.api.ph.service.PackageArtifactService;
import org.artifactory.core.api.ph.service.PackageRepoService;
import org.artifactory.core.api.ph.service.PackageUploadService;
import org.artifactory.core.api.ph.virtual.VirtualRepoTraversalBaseTest.TestingPackageMetadataRow;
import org.artifactory.core.api.ph.virtual.resolution.PackageMetadataRow;
import org.artifactory.core.api.ph.virtual.resolution.PackageMetadataVersionsSet;
import org.artifactory.core.api.ph.virtual.resolution.PackageRepo;
import org.artifactory.core.api.ph.virtual.resolution.PackageRequestContext;
import org.artifactory.core.api.ph.virtual.resolution.PackageVirtualController;
import org.artifactory.core.api.ph.virtual.resolution.PackageVirtualMerger;
import org.artifactory.core.api.ph.virtual.resolution.VirtualResolutionStrategyType;
import org.jetbrains.annotations.NotNull;
import org.jfrog.common.concurrency.JvmConflictsGuard;
import org.junit.jupiter.api.Assertions;
import org.mockito.Mock;
import org.mockito.Mockito;
import org.mockito.stubbing.Answer;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.testng.annotations.BeforeMethod;
import org.testng.annotations.DataProvider;
import org.testng.annotations.Test;

import java.io.ByteArrayInputStream;
import java.io.IOException;
import java.io.InputStream;
import java.nio.charset.StandardCharsets;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.concurrent.ConcurrentHashMap;
import java.util.concurrent.ExecutionException;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Executors;
import java.util.concurrent.Future;
import java.util.concurrent.TimeUnit;

import static org.junit.jupiter.api.Assertions.assertEquals;
import static org.mockito.ArgumentMatchers.any;
import static org.mockito.ArgumentMatchers.anyBoolean;
import static org.mockito.ArgumentMatchers.anyInt;
import static org.mockito.ArgumentMatchers.anyString;
import static org.mockito.ArgumentMatchers.eq;
import static org.mockito.ArgumentMatchers.refEq;
import static org.mockito.Mockito.doAnswer;
import static org.mockito.Mockito.times;
import static org.mockito.Mockito.verify;
import static org.mockito.Mockito.when;
import static org.mockito.MockitoAnnotations.openMocks;
import static org.testng.collections.Lists.newArrayList;


public class ConcurrentVirtualRepoTraversalTest {
    Logger log = LoggerFactory.getLogger(ConcurrentVirtualRepoTraversalTest.class);
    private static final String metadataPath = "a/b/metadata.test";
    private static final String virtualRepoKey = "virtualRepoKey";
    private static final String localRepoKey2 = "local_repo2";
    private static final String localRepoKey1 = "local_repo1";
    @Mock
    protected ConfProps confProps;

    @Mock
    protected PackageRepoService repoService;

    @Mock
    protected PackageArtifactService downloadService;
    @Mock
    protected PackageUploadService uploadService;

    @BeforeMethod
    public void setup() throws InterruptedException {
        openMocks(this);
    }

    /**
     * Test virtual repo resolution
     * 1. upload two different metadata files into two different repositories that belongs to virtual repo
     * 2. try to download metadata and expect to receive merged file.
     */
    @Test()
    public void testConcurrentMerge() throws PackageException, IOException {
        doAnswer(invocation -> 100).when(confProps).getIntValue(any(), anyInt());
        TestCache cache = new TestCache();
        List<? extends PackageRepo> subRepos = subRepositories(localRepoKey1, localRepoKey2);
        // Set sub repositories
        doAnswer(invocation -> subRepos).when(repoService).getSubRepositories(any(), any(), anyBoolean());
        when(repoService.getPackageType(virtualRepoKey)).thenReturn(PackageType.CRAN);
        //  Set sub-repository data (resource)
        initDownloadService(cache);
        // Update cache during the call to "uploadToVirtualCache"
        initUploadService(cache);
        // Implement virtual controller for the test
        PackageVirtualController<?> controller = controller(VirtualResolutionStrategyType.MERGER);
        int numberOfThreads = 50;
        ExecutorService executorService = Executors.newFixedThreadPool(numberOfThreads);
        List<Future<BaseResource>> futures = newArrayList();
        VirtualRepoTraversal virtualVisitor =
                new VirtualRepoTraversal(repoService, downloadService, uploadService,
                        new JvmConflictsGuard<>(100), confProps);
        try {
            executeConcurrentRequests(controller, numberOfThreads, executorService, futures, virtualVisitor);
            assertRequestResultIsValid(futures);
            assertOnlyOneMergeWasDone();
        } catch (Exception e) {
            log.error("Failed to execute task,", e);
            Assertions.fail();
        } finally {
            terminateExecutorService(executorService);
        }
    }

    private void initDownloadService(TestCache cache) throws IOException, PackageException {
        BaseResource result1 = new BaseResourceForTesting(localRepoKey1, metadataPath, null);
        BaseResource result2 = new BaseResourceForTesting(localRepoKey2, metadataPath, null);
        when(downloadService.getResource(refEq(new PackageDownloadContext(localRepoKey1, metadataPath))))
                .thenReturn(result1);
        when(downloadService.getResource(refEq(new PackageDownloadContext(localRepoKey2, metadataPath))))
                .thenReturn(result2);

        // Simulate cache (without the cache all threads will search in repositories and the test will fail)
        doAnswer((Answer<?>) invocationOnMock -> {
            String repoKey = invocationOnMock.getArgument(0);
            return cache.getResource(repoKey);
        }).when(downloadService).getResourceFromVirtualCache(any(), any());
    }

    private void initUploadService(TestCache cache) throws PackageException {
        doAnswer((Answer<?>) invocationOnMock -> {
            String repoKey = invocationOnMock.getArgument(0);
            String resourcePath = invocationOnMock.getArgument(1);
            InputStream inputStream = invocationOnMock.getArgument(2);
            BaseResource resource = new BaseResourceForTesting(repoKey, resourcePath, inputStream);
            cache.setResource(resource);
            return resource;
        }).when(uploadService).uploadToVirtualOrRemoteCache(any(), any(), any());
    }

    private void assertOnlyOneMergeWasDone() throws PackageException {
        verify(uploadService, times(1)).uploadToVirtualOrRemoteCache(eq(virtualRepoKey),
                eq(metadataPath), any());
    }

    private void assertRequestResultIsValid(List<Future<BaseResource>> futures)
            throws InterruptedException, ExecutionException, IOException {
        for (Future<BaseResource> future : futures) {
            BaseResource resource = future.get();
            assertEquals(virtualRepoKey, resource.getRepoKey());
            String data =
                    IOUtils.toString(((BaseResourceForTesting) resource).getInputStream(),
                            StandardCharsets.UTF_8);
            Assertions.assertTrue(data.contains("[local_repo2/a/b/metadata.test]"));
            Assertions.assertTrue(data.contains("[local_repo1/a/b/metadata.test]"));
        }
    }

    private void executeConcurrentRequests(PackageVirtualController<?> controller, int numberOfThreads,
            ExecutorService executorService, List<Future<BaseResource>> futures, VirtualRepoTraversal virtualVisitor) {
        for (int i = 0; i < numberOfThreads; i++) {
            futures.add(executorService.submit(() -> virtualVisitor
                    .resolve(new PackageDownloadContext(virtualRepoKey, metadataPath), controller,
                            (packageRequestContext, packageRepo) -> {
                                PackageRequestContext translated = null;
                                try {
                                    translated = controller
                                            .translateSubRepositoryContext(packageRepo, packageRequestContext);
                                    return downloadService.getResource(translated);
                                } catch (PackageException e) {
                                    if (translated == null) {
                                        translated = packageRequestContext;
                                    }
                                    return new PackageUnfoundedResource(translated.getRepoKey(), translated.getPath(),
                                            e.getStatus(),
                                            e.getPublicReason());
                                }
                            })));
        }
    }

    private void terminateExecutorService(ExecutorService executorService) {
        executorService.shutdown();
        try {
            boolean result = executorService.awaitTermination(500, TimeUnit.MILLISECONDS);
            log.debug("Executor service termination result:{}", result);
        } catch (InterruptedException e) {
            log.error("Failed to shout down executor service,", e);
            Assertions.fail();
        }
    }

    @DataProvider
    public Object[][] testMergerResolveLockTimeoutDataProvider() {
        return new Object[][] {
                {Map.of(), VirtualRepoTraversal.DEFAULT_MERGE_LOCK_TIMEOUT_SEC},
                {Map.of(VirtualRepoTraversal.MERGE_LOCK_TIMEOUT_PROP_GLOBAL_KEY, 400), 500},
                {Map.of(VirtualRepoTraversal.MERGE_LOCK_TIMEOUT_PROP_GLOBAL_KEY + ".cran", 600), 700},
                {Map.of(VirtualRepoTraversal.MERGE_LOCK_TIMEOUT_PROP_GLOBAL_KEY + ".cran." + virtualRepoKey,
                        800), 900}
        };
    }

    protected PackageRepo repository(String repoKey, boolean priority) {
        PackageRepo repo = Mockito.mock(PackageRepo.class);
        Mockito.when(repo.getKey()).thenReturn(repoKey);
        Mockito.when(repo.getFileContent(anyString())).thenReturn(null);
        Mockito.when(repo.isLocal()).thenReturn(true);
        Mockito.when(repo.isPriorityResolution()).thenReturn(priority);
        return repo;
    }


    @NotNull
    protected List<? extends PackageRepo> subRepositories(String... repoKeys) {
        List<PackageRepo> subRepositories = com.google.common.collect.Lists.newArrayList();
        for (String repoKey : repoKeys) {
            subRepositories.add(repository(repoKey, false));
        }
        return subRepositories;
    }


    @SuppressWarnings("rawtypes")
    protected PackageVirtualController controller(VirtualResolutionStrategyType strategy) throws PackageException {
        abstract class PackageVirtualControllerImpl implements PackageVirtualController {
        }

        PackageVirtualController controller = Mockito.mock(PackageVirtualControllerImpl.class,
                Mockito.CALLS_REAL_METHODS);
        Mockito.when(controller.getVirtualResolutionStrategy(any())).thenReturn(strategy);


        Mockito.doAnswer(invocation -> {
            BaseResource res = invocation.getArgument(0, BaseResource.class);
            String data = "[" + res.getRepoKey() + "/" + res.getPath() + "]";
            List<PackageMetadataRow> list = newArrayList(new TestingPackageMetadataRow(data));
            HashMap<String, List<PackageMetadataRow>> map = Maps.newHashMap();
            map.put(data, list);
            return map;
        }).when((PackageVirtualMerger) controller).extract(any(), any(), any(), any());
        //noinspection unchecked
        doAnswer(invocation -> {
            StringBuilder stringBuilder = new StringBuilder();
            Map<String, PackageMetadataVersionsSet<TestingPackageMetadataRow>> metadata = invocation.getArgument(0);
            for (PackageMetadataVersionsSet<TestingPackageMetadataRow> value : metadata.values()) {
                value.getVersions().forEach(row -> stringBuilder.append(row.getPackageName()));
            }
            return IOUtils.toInputStream(stringBuilder, StandardCharsets.UTF_8);
        }).when((PackageVirtualMerger) controller).toInputStream(any(), any());

        return controller;
    }

    private static class TestCache {
        private final Map<String, BaseResource> resourceMap = new ConcurrentHashMap<>();

        public void setResource(BaseResource resource) {
            resourceMap.put(resource.getRepoKey(), resource);
        }

        public BaseResource getResource(String repoKey) {
            return resourceMap.get(repoKey);
        }
    }

    protected static class BaseResourceForTesting implements BaseResource {
        private final long lastModified;
        private final String repoKey;
        private final String path;
        private final byte[] data;

        public BaseResourceForTesting(String repoKey, String path, InputStream inputStream) throws IOException {
            this.repoKey = repoKey;
            this.path = path;
            this.lastModified = System.currentTimeMillis();
            this.data = IOUtils.toByteArray(inputStream);
        }

        public InputStream getInputStream() {
            return new ByteArrayInputStream(data);
        }

        @Override
        public boolean isFound() {
            return true;
        }

        @Override
        public long getLastModified() {
            return lastModified;
        }

        @Override
        public boolean isExactQueryMatch() {
            return true;
        }

        @Override
        public String getRepoKey() {
            return repoKey;
        }

        @Override
        public String getPath() {
            return path;
        }

        @Override
        public boolean isExpired() {
            return false;
        }

        @Override
        public boolean isBlocked() {
            return false;
        }
    }
}
