package org.artifactory.storage.db.event.dao;

import org.apache.commons.compress.utils.Sets;
import org.artifactory.common.ConstantValues;
import org.artifactory.common.config.db.ArtifactoryDbProperties;
import org.artifactory.common.home.ArtifactoryHome;
import org.artifactory.common.property.ArtifactorySystemProperties;
import org.artifactory.storage.db.DbService;
import org.artifactory.storage.db.util.JdbcHelper;
import org.jfrog.storage.DbType;
import org.mockito.ArgumentCaptor;
import org.mockito.Mock;
import org.mockito.Mockito;
import org.mockito.MockitoAnnotations;
import org.testng.Assert;
import org.testng.annotations.AfterMethod;
import org.testng.annotations.BeforeMethod;
import org.testng.annotations.Test;

import java.sql.ResultSet;
import java.util.ArrayList;
import java.util.List;
import java.util.Set;
import java.util.concurrent.atomic.AtomicLong;
import java.util.stream.Collectors;

@Test
public class EventsDaoShiftEventsTest {

    private EventsDao eventsDao;
    @Mock
    private JdbcHelper jdbcHelper;
    @Mock
    private ArtifactoryDbProperties dbProperties;
    @Mock
    private ArtifactoryHome artifactoryHome;
    @Mock
    private ArtifactorySystemProperties artifactorySystemProperties;
    @Mock
    private DbService dbService;
    @Mock
    private ResultSet resultSet;

    private int eventsCount = (EventsDao.TMP_EVENTS_DELETE_BATCH_SIZE * 2) + 1;

    @BeforeMethod
    private void setup() throws Exception {
        MockitoAnnotations.openMocks(this);
        ArtifactoryHome.bind(artifactoryHome);
        Mockito.when(dbProperties.getDbType()).thenReturn(DbType.POSTGRESQL);
        Mockito.when(artifactoryHome.getArtifactoryProperties()).thenReturn(artifactorySystemProperties);
        Mockito.when(artifactorySystemProperties.getLongProperty(Mockito.eq(ConstantValues.shiftEventsSqlLimit)))
                .thenReturn(1000L);
        AtomicLong sequence = new AtomicLong(500);
        Mockito.when(dbService.nextId()).thenAnswer(invocation -> sequence.incrementAndGet());
        Mockito.when(jdbcHelper.executeSelect(Mockito.any(String.class))).thenReturn(resultSet);
        AtomicLong rsSize = new AtomicLong(eventsCount);

        Mockito.when(resultSet.next()).thenAnswer(rs -> rsSize.get() > 0);
        Mockito.when(resultSet.getLong(Mockito.eq("timestamp"))).thenReturn(System.currentTimeMillis());
        Mockito.when(resultSet.getInt(Mockito.eq("event_type"))).thenReturn(1);
        Mockito.when(resultSet.getString(Mockito.eq("path"))).thenReturn("repo1");
        Mockito.when(resultSet.getString(Mockito.eq("hints"))).thenReturn("hints");
        Mockito.when(resultSet.getLong(Mockito.eq("tmp_event_id"))).thenAnswer(invocation -> rsSize.getAndDecrement());
        createEventsDao();
    }

    @AfterMethod
    private void teardown() throws Exception {
        ArtifactoryHome.unbind();
    }

    @Test
    public void testDerbyEventIdSequence() throws Exception {
        Mockito.when(dbProperties.getDbType()).thenReturn(DbType.DERBY);
        createEventsDao();
        testShiftAndDeleteTmpEvents();
    }

    @Test
    public void testShiftAndDeleteTmpEvents() throws Exception {
        Set<String> repos = eventsDao.shiftAndDeleteTmpEvents();
        Assert.assertEquals(repos, Sets.newHashSet("repo1"));

        if (DbType.DERBY.equals(dbProperties.getDbType())) {
            ArgumentCaptor<Object> eventIdsCaptor = ArgumentCaptor.forClass(Object.class);

            Mockito.verify(dbService, Mockito.times(eventsCount)).nextId();

            Mockito.verify(jdbcHelper, Mockito.times(eventsCount))
                    .executeUpdate(Mockito.startsWith("INSERT INTO node_events"), Mockito.anyLong(), Mockito.anyInt(),
                            Mockito.anyString(), Mockito.anyString(), eventIdsCaptor.capture());

            Set<Long> eventIds =
                    eventIdsCaptor.getAllValues().stream().map(id -> (Long) id).collect(Collectors.toSet());
            for (int i = 1; i < eventsCount; i++) {
                Assert.assertTrue(eventIds.contains(500L + i), String.valueOf(500 + i));
            }
        } else {
            Mockito.verify(jdbcHelper, Mockito.times(eventsCount))
                    .executeUpdate(Mockito.startsWith("INSERT INTO node_events"), Mockito.anyLong(), Mockito.anyInt(),
                            Mockito.anyString(), Mockito.anyString());
        }
        ArgumentCaptor idsToDelete = ArgumentCaptor.forClass(Object[].class);
        Mockito.verify(jdbcHelper, Mockito.times(3))
                .executeUpdate(Mockito.startsWith("DELETE FROM node_events_tmp"), idsToDelete.capture());

        Assert.assertEquals(idsToDelete.getAllValues().size(), 3);

        Set<Long> deletedIds = (Set<Long>) idsToDelete.getAllValues().stream()
                .flatMap(item -> ((List<Long>) item).stream())
                .collect(Collectors.toSet());

        for (long id = eventsCount; id > 0; id--) {
            Assert.assertTrue(deletedIds.contains(id));
        }
    }

    @Test
    public void testPartitions() {
        List<List<Long>> partitions = new ArrayList<>();
        int count = (EventsDao.TMP_EVENTS_DELETE_BATCH_SIZE * 2) + 10;
        for (long i = 0; i < count; i++) {
            eventsDao.partition(partitions).add(i);
        }

        Assert.assertEquals(partitions.size(), 3);
        Assert.assertEquals(partitions.get(0).size(), EventsDao.TMP_EVENTS_DELETE_BATCH_SIZE);
        Assert.assertEquals(partitions.get(1).size(), EventsDao.TMP_EVENTS_DELETE_BATCH_SIZE);
        Assert.assertEquals(partitions.get(2).size(), 10);
    }

    private void createEventsDao() {
        eventsDao = new EventsDao(jdbcHelper);
        eventsDao.setDbProperties(dbProperties);
        eventsDao.setDbService(dbService);
        eventsDao.init();
    }
}
