package org.artifactory.storage.db.remotebintask.itest.dao;

import ch.qos.logback.classic.Level;
import com.jfrog.test.LogTestUtils;
import org.artifactory.addon.ArtifactoryRunningMode;
import org.artifactory.state.ArtifactoryServerState;
import org.artifactory.storage.db.itest.DbBaseTest;
import org.artifactory.storage.db.remotebintask.RemoteBinaryTaskDao;
import org.artifactory.storage.db.remotebintask.entity.RemoteBinaryTask;
import org.artifactory.storage.db.servers.dao.ArtifactoryServersDao;
import org.artifactory.storage.db.servers.model.ArtifactoryServer;
import org.artifactory.storage.db.servers.model.ArtifactoryServerRole;
import org.artifactory.storage.db.util.JdbcHelper;
import org.jetbrains.annotations.NotNull;
import org.springframework.beans.factory.annotation.Autowired;
import org.testng.Assert;
import org.testng.annotations.AfterMethod;
import org.testng.annotations.BeforeMethod;
import org.testng.annotations.DataProvider;
import org.testng.annotations.Test;

import java.util.Arrays;
import java.util.List;
import java.util.concurrent.TimeUnit;
import java.util.concurrent.atomic.AtomicLong;
import java.util.stream.Collectors;

import static org.artifactory.storage.db.remotebintask.RemoteBinaryTaskDao.COLUMNS;
import static org.artifactory.storage.db.remotebintask.RemoteBinaryTaskDao.TABLE_NAME;
import static org.artifactory.storage.db.remotebintask.entity.RemoteBinaryTask.RemoteBinaryTaskBuilder;
import static org.artifactory.storage.db.remotebintask.entity.RemoteBinaryTask.builder;

@Test
public class RemoteBinaryTaskDaoITest extends DbBaseTest {


    @Autowired
    private RemoteBinaryTaskDao dao;

    @Autowired
    private ArtifactoryServersDao serversDao;

    @Autowired
    private JdbcHelper jdbcHelper;

    AtomicLong nextId = new AtomicLong();


    @Test
    public void testAddTask() throws Exception {
        long time = System.currentTimeMillis();
        dao.add(nextId.incrementAndGet(), "sha1",
                "repo1", "http://host1:8081/artifactory/repo1");
        List<RemoteBinaryTask> tasks = dao.getAll();
        Assert.assertEquals(tasks.size(), 1);
        Assert.assertTrue(tasks.get(0).getId() > 0);
        Assert.assertEquals(tasks.get(0).getCount(), 0);
        Assert.assertEquals(tasks.get(0).getLocalRepoKey(), "repo1");
        Assert.assertEquals(tasks.get(0).getRemoteUrlHash(),
                "http://host1:8081/artifactory/repo1".hashCode());
        Assert.assertEquals(tasks.get(0).getSha1(), "sha1");
        Assert.assertNull(tasks.get(0).getServerId());
        Assert.assertNull(tasks.get(0).getStartTime());
        Assert.assertTrue(tasks.get(0).getTime() >= time);
    }

    // test that converted task without url will be readable
    @Test
    public void testGetTaskWIthNullRemoteUrl() throws Exception {
        dao.add(nextId.incrementAndGet(), "sha1", "repo1", null);
        List<RemoteBinaryTask> tasks = dao.getAll();
        Assert.assertEquals(tasks.get(0).getRemoteUrlHash(), 0);
    }

    @Test
    public void testAllocate() throws Exception {
        dao.add(nextId.incrementAndGet(), "sha1", "repo1", "http://host1:8081/artifactory/repo1");
        RemoteBinaryTask taskBefore = dao.getAll().get(0);

        long startTime = System.currentTimeMillis();
        dao.allocate(taskBefore.getId(), "serverId");

        RemoteBinaryTask taskAfter = dao.getAll().get(0);
        Assert.assertEquals(taskAfter.getId(), taskBefore.getId());
        Assert.assertEquals(taskAfter.getTime(), taskBefore.getTime());
        Assert.assertEquals(taskAfter.getServerId(), "serverId");
        Assert.assertEquals(taskAfter.getSha1(), taskBefore.getSha1());
        Assert.assertEquals(taskAfter.getRemoteUrlHash(), taskBefore.getRemoteUrlHash());
        Assert.assertTrue(taskAfter.getStartTime() >= startTime);
        Assert.assertEquals(taskAfter.getCount(), taskBefore.getCount() + 1);
    }

    @Test
    public void testDeallocateById() throws Exception {
        dao.add(nextId.incrementAndGet(),
                "sha1", "repo1", "http://host1:8081/artifactory/repo1");
        RemoteBinaryTask taskBeforeAllocate = dao.getAll().get(0);

        dao.allocate(taskBeforeAllocate.getId(), "serverId");
        RemoteBinaryTask taskAfterAllocate = dao.getAll().get(0);

        dao.deAllocateById(taskAfterAllocate.getId());
        RemoteBinaryTask taskAfterDeallocate = dao.getAll().get(0);

        Assert.assertEquals(taskAfterDeallocate.getId(), taskBeforeAllocate.getId());
        Assert.assertEquals(taskAfterDeallocate.getTime(), taskBeforeAllocate.getTime());
        Assert.assertNull(taskAfterDeallocate.getServerId());
        Assert.assertEquals(taskAfterDeallocate.getSha1(), taskBeforeAllocate.getSha1());
        Assert.assertEquals(taskAfterDeallocate.getRemoteUrlHash(), taskBeforeAllocate.getRemoteUrlHash());
        Assert.assertNull(taskAfterDeallocate.getStartTime());
        Assert.assertEquals(taskAfterDeallocate.getCount(), taskAfterAllocate.getCount());
    }

    @Test
    public void testDeallocateByServerId() throws Exception {
        dao.add(nextId.incrementAndGet(),
                "sha11", "repo1", "http://host1:8081/artifactory/repo1");
        dao.add(nextId.incrementAndGet(),
                "sha12", "repo1", "http://host1:8081/artifactory/repo1");
        dao.add(nextId.incrementAndGet(),
                "sha13", "repo1", "http://host1:8081/artifactory/repo1");

        List<RemoteBinaryTask> tasksBeforeAllocate = dao.getAll();

        dao.allocate(tasksBeforeAllocate.get(0).getId(), "server1");
        dao.allocate(tasksBeforeAllocate.get(1).getId(), "server2");
        dao.allocate(tasksBeforeAllocate.get(2).getId(), "server1");

        List<RemoteBinaryTask> tasksAfterAllocate = dao.getAll();
        List<RemoteBinaryTask> expectedDeallocatedTasks = tasksAfterAllocate.stream()
                .filter(t -> t.getServerId().equals("server1")).collect(Collectors.toList());
        List<RemoteBinaryTask> expectedAllocatedTasks = tasksAfterAllocate.stream()
                .filter(t -> t.getServerId().equals("server2")).collect(Collectors.toList());

        dao.deAllocateByServerId("server1");

        List<RemoteBinaryTask> tasksAfterDeallocate = dao.getAll();


        for (RemoteBinaryTask task : expectedAllocatedTasks) {
            RemoteBinaryTask taskAfterDeallocate = tasksAfterDeallocate.stream().filter(t -> t.getId() == task.getId())
                    .findFirst().orElseThrow();

            Assert.assertEquals(taskAfterDeallocate.getServerId(), task.getServerId());
            Assert.assertEquals(taskAfterDeallocate.getStartTime(), task.getStartTime());

            Assert.assertEquals(taskAfterDeallocate.getId(), task.getId());
            Assert.assertEquals(taskAfterDeallocate.getTime(), task.getTime());
            Assert.assertEquals(taskAfterDeallocate.getSha1(), task.getSha1());
            Assert.assertEquals(taskAfterDeallocate.getRemoteUrlHash(), task.getRemoteUrlHash());
            Assert.assertEquals(taskAfterDeallocate.getCount(), task.getCount());
        }

        for (RemoteBinaryTask task : expectedDeallocatedTasks) {
            RemoteBinaryTask taskAfterDeallocate = tasksAfterDeallocate.stream().filter(t -> t.getId() == task.getId())
                    .findFirst().orElseThrow();

            Assert.assertNull(taskAfterDeallocate.getServerId());
            Assert.assertNull(taskAfterDeallocate.getStartTime());

            Assert.assertEquals(taskAfterDeallocate.getId(), task.getId());
            Assert.assertEquals(taskAfterDeallocate.getTime(), task.getTime());
            Assert.assertEquals(taskAfterDeallocate.getSha1(), task.getSha1());
            Assert.assertEquals(taskAfterDeallocate.getRemoteUrlHash(), task.getRemoteUrlHash());
            Assert.assertEquals(taskAfterDeallocate.getCount(), task.getCount());
        }
    }

    @Test
    public void testRemove() throws Exception {
        dao.add(nextId.incrementAndGet(),
                "sha11", "repo1", "http://host1:8081/artifactory/repo1");

        List<RemoteBinaryTask> tasksBeforeRemove = dao.getAll();

        dao.remove(tasksBeforeRemove.get(0).getId());

        Assert.assertNull(dao.getById(tasksBeforeRemove.get(0).getId()));
    }

    @Test
    public void testGetAll() throws Exception {
        dao.add(nextId.incrementAndGet(),
                "sha11", "repo1", "http://host1:8081/artifactory/repo1");
        dao.add(nextId.incrementAndGet(),
                "sha12", "repo1", "http://host1:8081/artifactory/repo1");

        List<RemoteBinaryTask> tasks = dao.getAll();
        Assert.assertEquals(tasks.size(), 2);
    }

    @Test
    public void testGetById() throws Exception {
        dao.add(nextId.incrementAndGet(),
                "sha11", "repo1", "http://host1:8081/artifactory/repo1");
        dao.add(nextId.incrementAndGet(),
                "sha12", "repo1", "http://host1:8081/artifactory/repo1");

        List<RemoteBinaryTask> tasks = dao.getAll();
        RemoteBinaryTask task = dao.getById(tasks.get(0).getId());
        Assert.assertNotNull(task);
        Assert.assertEquals(task.getId(), tasks.get(0).getId());
    }

    @DataProvider
    private Object[][] testNextDataProvider() throws Exception {
        artifactoryHomeBoundTest.bindArtifactoryHome();
        try {
            ArtifactoryServer offline = new ArtifactoryServer("server1", 1, "contextUrl", 1111,
                    ArtifactoryServerState.OFFLINE, ArtifactoryServerRole.MEMBER, 0, "7.21.1", 0,
                    0, ArtifactoryRunningMode.HA, "licenceKey");
            ArtifactoryServer online = new ArtifactoryServer("server1", 1, "contextUrl", 1111,
                    ArtifactoryServerState.RUNNING, ArtifactoryServerRole.MEMBER,
                    System.currentTimeMillis() + TimeUnit.MINUTES.toMillis(10),
                    "7.21.1", 0, 0, ArtifactoryRunningMode.HA, "licenceKey");

            RemoteBinaryTaskBuilder task = builder()
                    .id(1)
                    .time(0);

            return new Object[][] {
                    //unallocated task, count >= max count ==> Null
                    {Arrays.asList(task.count(10).build()), Arrays.asList(online), null},

                    //unallocated task, count < max count  ==> Task
                    {Arrays.asList(task.count(0).build()), Arrays.asList(online), 1L},

                    //unallocated task, count < max count, retry time not now ==> Null
                    {Arrays.asList(task.count(9).time(System.currentTimeMillis())
                            .build()), Arrays.asList(online), null},

                    //unallocated task, count < max count, retry time now ==> Task
                    {Arrays.asList(task.count(9).time(0)
                            .build()), Arrays.asList(online), 1L},

                    //allocated task, exec duration not passed ==> Null
                    {Arrays.asList(task.count(1).time(0)
                            .serverId("server1").startTime(System.currentTimeMillis())
                            .build()), Arrays.asList(online), null},

                    //allocated task, exec duration not passed, server offline ==> Task
                    {Arrays.asList(task.count(1).time(0)
                            .serverId("server1").startTime(System.currentTimeMillis())
                            .build()), Arrays.asList(offline), 1L},

                    //allocated task, exec duration passed, server online ==> Task
                    {Arrays.asList(task.count(1).time(0)
                            .serverId("server1").startTime(0L)
                            .build()), Arrays.asList(online), 1L},
            };
        } finally {
            artifactoryHomeBoundTest.unbindArtifactoryHome();
        }
    }

    @AfterMethod
    @BeforeMethod
    private void cleanup() throws Exception {
        jdbcHelper.executeUpdate("DELETE FROM " + TABLE_NAME);
        jdbcHelper.executeUpdate("DELETE FROM artifactory_servers");
        LogTestUtils.setLoggingLevel(org.jfrog.storage.JdbcHelper.class, Level.ALL);
    }

    @Test(dataProvider = "testNextDataProvider", alwaysRun = true)
    public void testNext(List<RemoteBinaryTask> tasks, List<ArtifactoryServer> servers, Long expectedTaskId)
            throws Exception {
        for (RemoteBinaryTask task : tasks) {
            addTask(task);
        }

        for (ArtifactoryServer server : servers) {
            serversDao.createArtifactoryServer(server);
        }

        List<RemoteBinaryTask> tasksRes = dao.getTasks("server1", 1, 10,
                30000, 60000);
        if (expectedTaskId != null) {
            Assert.assertNotNull(tasksRes.get(0));
            Assert.assertEquals(tasksRes.get(0).getId(), expectedTaskId.longValue());
        } else {
            Assert.assertEquals(tasksRes.size(), 0);
        }
    }

    @Test
    public void testGetMultipleTasks()
            throws Exception {
        RemoteBinaryTaskBuilder builder = builder();
        RemoteBinaryTask t1 = builder.id(1).serverId("sid").sha1("sha1").count(0).build();
        RemoteBinaryTask t2 = builder.id(2).serverId("sid").sha1("sha1").count(0).build();
        RemoteBinaryTask t3 =
                builder.id(3).serverId("sid").sha1("sha1").count(0).time(System.currentTimeMillis()).build();
        RemoteBinaryTask t4 = builder.id(4).serverId("sid").sha1("sha1").count(11).build();
        addTask(t1);
        addTask(t2);
        addTask(t3);
        addTask(t4);

        List<RemoteBinaryTask> tasksRes = dao.getTasks("server1", 5, 10,
                30000, 60000);
        Assert.assertNotNull(tasksRes);
        Assert.assertEquals(tasksRes.size(), 3);

    }

    @Test
    public void testAllocatedTaskByApiNotPicked() throws Exception {
        ArtifactoryServer online = getOnlineServer();
        serversDao.createArtifactoryServer(online);
        dao.add(nextId.incrementAndGet(), "sha1", "server1", "");
        List<RemoteBinaryTask> tasksRes = dao.getTasks("server1", 1, 10,
                30000, 60000);

        List<RemoteBinaryTask> tasksRes2 = dao.getTasks("server1", 1, 10,
                30000, 60000);

        Assert.assertNotNull(tasksRes);
        Assert.assertEquals(tasksRes.size(), 1);
        Assert.assertNotNull(tasksRes2);
        Assert.assertEquals(tasksRes2.size(), 0);

    }

    @NotNull
    private ArtifactoryServer getOnlineServer() {
        return new ArtifactoryServer("server1", 1, "contextUrl", 1111,
                ArtifactoryServerState.RUNNING, ArtifactoryServerRole.MEMBER,
                System.currentTimeMillis() + TimeUnit.MINUTES.toMillis(10),
                "7.21.1", 0, 0, ArtifactoryRunningMode.HA, "licenceKey");
    }

    private void addTask(RemoteBinaryTask task) throws Exception {
        //id, sha1, time, local_repo_key, remote_url_hash, start_time, count, server_id
        jdbcHelper.executeUpdate("INSERT INTO " + TABLE_NAME + "(" + COLUMNS + ") " +
                        "VALUES (?, ?, ?, ?, ?, ?, ?, ?)",
                task.getId(), "sha1111", task.getTime(), "repo1", "remoteUrl".hashCode(), task.getStartTime(),
                task.getCount(), task.getServerId());
    }

}
